unit PairingU;
{The TPairing component is used to support a many-many relationship. The
component contains a list of pairs of record identifiers (we call them rids).
The inclusion of a pair (A,B) on the list means that the record with rid A
in the first file is somehow related to the record with rid B in the second
file. For example, there may be a file of authors and a file of books. The
pair (A,B) mean that the author with rid A has written or cowritten the book
with rid B. Each book may have several authors and each author may have
written several books. TPairings keep track of these kinds of facts. There
are routines that add pairs, delete pairs, find one part of a pair given
the other part and reading or storing TPairings to or from disk datasets.}

interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  PairItem = record
    A: longint;
    B: longint;
    end;
  PairArray = array [0..8000] of PairItem;
  PairArrayPtr = ^PairArray;

  RidArray = array [0..16000] of longint;
  RidArrayPtr = ^RidArray;

  TPairing = class(TComponent)
  private
    Capacity: cardinal;
    FCount: cardinal;
    PL: PairArrayPtr;
    FModified: boolean;
    FFileName: string;
    F: file of PairItem;
    { Private declarations }
  protected
    procedure SetCapacity(Value: cardinal);
    procedure Grow;
    { Protected declarations }
  public
    destructor Destroy; override;
    property Count: cardinal
      read FCount;
    property Modified: boolean
      read FModified;
    { Public declarations }
  published
    { Published declarations }
    function PairByIndex(j: cardinal): PairItem;
    function Exist(AA,BB: longint): boolean; {Is (AA,BB) on the list?}
    {The next sets a pointer to a RidArray that contains all the
      B's associated with A=AA. The 0th item is the number on the list}
    procedure BsOf(AA: longint; var RAP: RidArrayPtr);
    {Similar to the above}
    procedure AsOf(BB: longint; var RAP: RidArrayPtr);
    procedure Add(AA,BB: longint);  {Add the pair (AA,BB)}
    procedure Delete(AA,BB: longint);  {Delete the pair (AA,BB)}
    procedure DeleteAllA(AA: longint); {Delete all pairs (AA,X) for all X}
    procedure DeleteAllB(BB: longint); {Delete all pairs (X,BB) for all X}
    procedure LoadFromFile;
    procedure SaveToFile;
    property FileName: string
      read  FFileName
      write FFileName;
  end;

procedure Register;

procedure Reclaim(var RAP: RidArrayPtr);

implementation

procedure Register;
begin
  RegisterComponents('Len''s', [TPairing]);
end;


procedure Reclaim(var RAP: RidArrayPtr);

begin
  if RAP = nil then exit;
  FreeMem(RAP);
{  if RAP^[0] = 0 then FreeMem(RAP,40)
  else FreeMem(RAP,40*((RAP^[0] div 10)+1)); }
  RAP := nil;
  end;


destructor TPairing.Destroy;

begin
  SetCapacity(0);
  inherited Destroy;
  end;


procedure TPairing.SetCapacity(Value: cardinal);
{Sets the capacity of the TPairing}

var
  P: PairArrayPtr;

begin
  if Value = Capacity then exit;
  if Value = 0 then P := nil
  else begin
    GetMem(P,8*Value);
    if FCount <> 0 then move(PL^,P^,8*FCount);
    end;
  if Capacity <> 0 then FreeMem(PL);
  PL := P;
  Capacity := Value;
  end;


procedure TPairing.Grow;

begin
  SetCapacity(Capacity+16);
  end;


function TPairing.PairByIndex(j: cardinal): PairItem;

begin
  if (j < 0) or (j > FCount) then begin
    ShowMessage('Index out of bounds');
    exit;
    end;
  Result := PL^[j-1];
  end;


function TPairing.Exist(AA,BB: longint): boolean;

var
  j: cardinal;

begin
  Result := FALSE; {Assume not on list}
  for j := 1 to FCount do
    if (PL^[j-1].A = AA) and (PL^[j-1].B = BB) then begin
      Result := TRUE;
      break;
      end;
  end;



procedure InsertRid(var R: RidArrayPtr; X: longint);
{This useful procedure adds the rid X to the RidArray pointed to by R.
If the RidArray needs to be expanded this will do it and adjust the
size as well.}

var
  T: RidArrayPtr;

begin
  if (R^[0] mod 10 = 9) then begin
    GetMem(T,4*(R^[0]+10)+4);
    move(R^,T^,4*R^[0]+4);
    FreeMem(R);
    R := T;
    end;
  inc(R^[0]);
  R^[R^[0]] := X;
  end;



procedure TPairing.BsOf(AA: longint; var RAP: RidArrayPtr);

var
  j: cardinal;

begin
  GetMem(RAP,40);
  RAP^[0] := 0;
  if FCount = 0 then exit;
  for j := 1 to FCount do
    if PL^[j-1].A = AA then InsertRid(RAP,PL^[j-1].B);
  end;



procedure TPairing.AsOf(BB: longint; var RAP: RidArrayPtr);

var
  j: cardinal;

begin
  GetMem(RAP,40);
  RAP^[0] := 0;
  if FCount = 0 then exit;
  for j := 1 to FCount do
    if PL^[j-1].B = BB then InsertRid(RAP,PL^[j-1].A);
  end;


procedure TPairing.Add(AA,BB: longint);
{Add a pair at the end of the list}

begin
  if Exist(AA,BB) then exit;
  if Capacity = FCount then Grow;
  inc(FCount);
  PL^[FCount-1].A := AA;
  PL^[FCount-1].B := BB;
  FModified := TRUE;
  end;


procedure TPairing.Delete(AA,BB: longint);
{Copy the last pair to the pair with rid r effectively deleting it}

var
  Kill: cardinal;

begin
  if FCount = 0 then exit;
  Kill := 0;
  while (Kill < FCount) and
    ((PL^[Kill].A <> AA) or (PL^[Kill].B <> BB)) do inc(Kill);
  if (Kill < FCount) then begin
    PL^[Kill] := PL^[FCount-1];
    dec(FCount);
    FModified := TRUE;
    end;
  end;



procedure TPairing.DeleteAllA(AA: longint);
{Delete all pairs with A rid of AA}

var
  j: cardinal;

begin
  if FCount = 0 then exit;
  j := 0;
  repeat
    while (j < FCount) and (PL^[j].A <> AA) do inc(j);
    if (j < FCount) then begin
      PL^[j] := PL^[FCount-1];
      dec(FCount);
      FModified := TRUE;
      end;
    until j >= FCount;
  end;



procedure TPairing.DeleteAllB(BB: longint);
{Delete all pairs with B rid of BB}

var
  j: cardinal;

begin
  if FCount = 0 then exit;
  j := 0;
  repeat
    while (j < FCount) and (PL^[j].B <> BB) do inc(j);
    if (j < FCount) then begin
      PL^[j] := PL^[FCount-1];
      dec(FCount);
      FModified := TRUE;
      end;
    until j >= FCount;
  end;



procedure TPairing.LoadFromFile;
{Read the pairs from an external file, if there is one.
Otherwise it sets the capacity to 10 assuming a new
TPairing not previously stored.}


var
  j,c: cardinal;

begin
  FModified := FALSE;
  if FileName = '' then begin
    Capacity := 10;
    exit;
    end;
  AssignFile(F,FileName);
  {$I-} reset(F); {$I+}
  if IOResult <> 0 then begin
    Capacity := 10;
    exit;
    end;
  FCount := 0; {This makes SetCapacity work OK}
  c := FileSize(F);
  SetCapacity(c+16);
  FCount := c;
  for j := 1 to FCount do read(F,PL^[j-1]);
  CloseFile(F);
  end;



procedure TPairing.SaveToFile;
{Write the pairs to an external file}

var
  j: cardinal;

begin
  if FFileName = '' then begin
    ShowMessage('File not named');
    exit;
    end;
  AssignFile(F,FFileName);
  rewrite(F);
  for j := 1 to FCount do write(F,PL^[j-1]);
  CloseFile(F);
  FModified := FALSE;
  end;

end.
