unit RidKeyU;
{A RidKey is a pair one member of which is a longint and the other a string.
The string is meant to be a key to a record and the longint is meant to be
the record identifier. There is a one-one correspondence between strings and
longints. A given key can have only one associated longint and a given
longint can have only one string. The unit contains routines that return
the key string given the record identifier and the record identifier given
a key. There are other routines that add pairs, delete them and read or store
them in disk data sets. The maximum length of strings is given by a property.}

interface

uses Classes,
     VarFileU;

type
  RidKeyItem = record
    rid: longint;
    key: ^ShortString;
    end;
  RidKeyArray = array[0..0] of RidKeyItem;
  RidKeyArrayPtr = ^RidKeyArray;

  PickStringProc = function(Buf: pointer): ShortString;

  TRidKey = class (TComponent)
  private
    FStringLength: cardinal; {Maximum length of a string}
    FCapacity: cardinal; {The number of pairs possible}
    FCount: cardinal;   {The number of pairs now}
    FModified: boolean; {TRUE if any pairs are added or deleted}
    RKL: RidKeyArrayPtr; {Points to the array of pairs in memory}
    Kill: cardinal; {Set by KeyOf and RidOf in case of delete.}
    FFileName: string;  {Name of external data set to hold RidKey}
    RKF: TextFile;
  protected
    procedure SetStringLength(Value: cardinal);
    procedure SetCapacity(Value: cardinal);
    procedure Grow;       {Called when needed}
  public
    destructor Destroy; override;
    property Count: cardinal
      read  FCount;
    property Modified: boolean
      read  FModified;
  published
    function RidOf(s: shortString): longint;
    function KeyOf(r: longint): ShortString;
    function KeyByIndex(j:cardinal): ShortString;
    procedure Add(r: longint; s: ShortString);
    procedure DeleteR(r: longint);
    procedure DeleteS(s: ShortString);
    procedure LoadFromFile;
    procedure SaveToFile;
    procedure ExtractFromFile(var VF: TVarFile; Pick: PickStringProc);
    property FileName: string
      read  FFileName
      write FFileName;
    property StringLength: cardinal
      read  FStringLength
      write SetStringLength;
    property Capacity: cardinal
      read  FCapacity
      write SetCapacity;
    end;


procedure Register;



implementation

uses Sysutils, Dialogs;


procedure Register;

begin
  RegisterComponents('Len''s',[TRidKey]);
  end;


procedure TRidKey.SetStringLength(Value: cardinal);

begin
  if Value = FStringLength then exit;
  FStringLength := Value;
  end;


destructor TRidKey.Destroy;

begin
  Capacity := 0;
  inherited Destroy;
  end;


procedure TRidKey.SetCapacity(Value: cardinal);
{Changes the capacity of the RidKey in memory}

var
  P: RidKeyArrayPtr;
  j: cardinal;

begin
  if Value = FCapacity then exit;
  if Value = 0 then begin
    P := nil;
    for j := 1 to FCount do
      FreeMem(RKL^[j-1].key);
    end
  else begin
    GetMem(P,8*Value);
    if FCount <> 0 then move(RKL^,P^,8*FCount);
    end;
  if FCapacity <> 0 then FreeMem(RKL);
  RKL := P;
  FCapacity := Value;
  end;


procedure TRidKey.Grow;

begin
  Capacity := Capacity + 16;
  end;


function TRidKey.RidOf(s: ShortString): longint;
{Returns the  rid of the given string s or -1 if the string
does not exist. }

begin
  Result := -1; {Assume not there}
  if FCount = 0 then exit;
  Kill := 0;
  while (Kill < FCount) and (RKL^[Kill].key^ <> s) do inc(Kill);
  if (Kill < FCount) and (RKL^[Kill].key^ = s) then
     Result := RKL^[Kill].rid;
  end;


function TRidKey.KeyOf(r: longint): ShortString;
{Returns the string for a given rid or NULL if the rid does not exist}

begin
  Result := '';  {Assume not there}
  Kill := 0;
  while (Kill < FCount) and (RKL^[Kill].rid <> r) do inc(Kill);
  if Kill < FCount then Result := RKL^[Kill].key^;
  end;


function TRidKey.KeyByIndex(j: cardinal): ShortString;

begin
  if (j > FCount) or (j < 1) then begin
    ShowMessage('Index out of bounds');
    exit;
    end;
  Result := RKL^[j-1].key^;
  end;



procedure TRidKey.Add(r: longint; s: ShortString);
{Add a pair at the end of the list}

begin
  if KeyOf(r) <> '' then begin
    ShowMessage('String already on list');
    exit;
    end;
  if RidOf(s) <> -1 then begin
    ShowMessage('Record id already on list');
    exit;
    end;
  if FCapacity = FCount then Grow;
  inc(FCount);
  RKL^[FCount-1].rid := r;
  GetMem(RKL^[FCount-1].key,FStringLength+1);
  RKL^[FCount-1].key^ := s;
  FModified := TRUE;
  end;


procedure TRidKey.DeleteR(r: longint);
{Copy the last pair to the pair with rid r effectively deleting it}

begin
  if KeyOf(r) = '' then exit; {Pair absent already.}
  FreeMem(RKL^[Kill].key);
  RKL^[Kill] := RKL^[FCount-1];
  dec(FCount);
  FModified := TRUE;
  end;


procedure TRidKey.DeleteS(s: ShortString);
{Copy the last pair to the pair with string s effectively deleting it.}

begin
  if RidOf(s) = -1 then exit; {Pair absent already}
  FreeMem(RKL^[Kill].key);
  RKL^[Kill] := RKL^[FCount-1];
  dec(FCount);
  FModified := TRUE;
  end;



procedure TRidKey.LoadFromFile;
{Read the pairs from an external file, if there is one.
Otherwise it sets the capacity to 10 assuming a new
RidKey not previously stored.}


var
  j,c: cardinal;
  s: ShortString;

begin
  FModified := FALSE;
  if FileName = '' then begin
    Capacity := 16;
    exit;
    end;
  AssignFile(RKF,FileName);
  {$I-} reset(RKF); {$I+}
  if IOResult <> 0 then begin
    Capacity := 16;
    exit;
    end;
  FCount := 0; {This makes SetCapacity work OK}
  Readln(RKF,s);
  c := StrToInt(s);
  Capacity := c+16;
  FCount := c;
  for j := 1 to FCount do begin
    Readln(RKF,s);
    RKL^[j-1].rid := StrToInt(s);
    GetMem(RKL^[j-1].key,FStringLength+1);
    Readln(RKF,RKL^[j-1].key^);
    end;
  CloseFile(RKF);
  end;



procedure TRidKey.SaveToFile;
{Write the pairs to an external file}

var
  j: cardinal;

begin
  if FFileName = '' then begin
    ShowMessage('File not named');
    exit;
    end;
  AssignFile(RKF,FFileName);
  rewrite(RKF);
  writeln(RKF,IntToStr(FCount));
  for j := 1 to FCount do begin
    writeln(RKF,IntToStr(RKL^[j-1].rid));
    writeln(RKF,RKL^[j-1].key^);
    end;
  CloseFile(RKF);
  FModified := FALSE;
  end;


procedure TRidKey.ExtractFromFile(var VF: TVarFile; Pick: PickStringProc);

var
  j,BS: cardinal;
  Buf: pointer;

begin
  VF.Open;
  Capacity := VF.MaxRec+16;
  for j := 1 to VF.MaxRec do begin
    BS := VF.GetBufSize(j);
    if BS <> 0 then begin
      GetMem(Buf,BS);
      VF.Fetch(j,Buf);
      Add(j,Pick(Buf));
      FreeMem(Buf);
      end;
    end;
  end;


end.
