unit VarFileU;
{This unit contains all the routines needed to handle files with
records of varying lengths. There are routines that insert new
records, fetch, restore and delete old ones, and a few housekeeping
routines.

The records in the file occupy groups of 'contiguous blocks'. All
blocks have the same size. The block size is determined at the
creation of the file and cannot be changed. The reason for the
block size is to allow files whose records are variable numbers of
same sized items; for example, a bunch of longints, or a bunch of
string[6]'s. Records are transmitted from and to the file in
groups of blocks.

The structure of the file is as follows:

The 32 byte Prefix as described below. Then comes a series of LocItems
that locate the records in the file. The LocItems contain two fields:
the block number where the record starts and the count field which tells
how many blocks are in the record. As the records change size they move
around the file and this list of LocItems needs to be updated frequently.
The number of LocItems is given by the RecCapacity field of the Prefix.
The RecCapacity is determined when the file is created but can be
changed later by the VarFileUtility that comes in the VarFile package.
Two other fields in the Prefix that handle the location of records are
the RecCount, the actual number of records in the file now; and MaxRec
the maximum number of records ever in the file. MaxRec is handy for
estimating how big to make RecCapacity when the VarFileUtility program
reconstitutes the file.

Then comes a series of LocItems that locate free areas in the file.
As records shrink or grow the places they used to occupy no longer hold
them and become 'free'. These free areas can be used for other records
and routines always look for a free area first before enlarging the data
set for storing new or changed records. The number of LocItems in the
free area is given by the FreeCapacity field of the Prefix. It is
determined when the file is created but can also be changed by the
VarFileUtility program. The fields FreeCount and MaxFree are similar to
the same fields for locator LocItems.

Then the records begin with the first block (block 0) beginning
immediately after the free list. Records are identified by record
identifiers (rid's) that are positive longints. The field NextRid
of the Prefix contains the next rid to be used.  There are two methods
that delete records. One just deletes the record but the other also builds
a stack in the locator list of deleted record numbers. When new records
are inserted this stack is consulted and, if not empty, the most recently
pushed will be returned as the next record identifier to be used.}


interface

uses Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs;

type
  LocItem = record
    start:  longint;     {The block number where the record starts}
    count:  longint;     {The number of blocks in the record}
    end;

  Prefix = record
    RecCapacity:  longint;  {Number of records possible.}
    RecCount:     longint;  {Number of records now}
    MaxRec:       longint;  {Largest number of records at any time.}
    FreeCapacity: longint;  {Number allowed on free list}
    FreeCount:    longint;  {Number of free list now}
    MaxFree:      longint;  {Largest number on free list at any time}
    BlockSize:    longint;  {Number of bytes in a block}
    NextRid:      longint;  {Record id of next record to be inserted}
    end;

  LocList = array [0..0] of LocItem;
  LocListPtr = ^LocList;

  FlushType = (Often, Never);

  TVarFile = class(TComponent)
  private
    { Private declarations }
    P:             Prefix;
    LI:            LocItem;  {Filled when record existence is tested}
    FOpen:         boolean;    {TRUE when file is open}
    FFlush:        FlushType;   {value Never flushes only at Close}
    RLP:           LocListPtr;  {Points to Locator LocItems}
    FLP:           LocListPtr;  {Points to Free List LocItems}
    RecOffset:     longint;     {Byte offset to first byte of data}
    FFileName:     string;
    F:             file;
    procedure  SetFileName(Value: string);
    function   FindArea(BlocksWanted: longint): longint;
    procedure  AddToFreeList(GivenStart, GivenCount: longint);
    procedure  DoFlush;
    function  DoDelete(rid: longint; Control: cardinal): boolean;
  protected
    { Protected declarations }
  public
    { Public declarations }
    function    PointToPrefix: pointer;
    property  RecCount: longint
       read P.RecCount;
    property  MaxRec: longint
       read P.MaxRec;
  published
    { Published declarations }
    constructor   Create(AOwner: TComponent); override;
    destructor    Destroy; override;
    procedure     Open;
    procedure     Close;
    function      Exist(rid: longint): boolean;
    function      GetBufSize(rid: longint): longint;
    function      Fetch(rid: longint; Buf: pointer): boolean;
    function      GetARid: longint;
    function      Insert(rid: longint; Buf: pointer;
                          RL: longint): boolean;
    function      Restore(rid: longint; Buf: pointer;
                          RL: longint): boolean;
    function      Delete(rid: longint): boolean;
    function      DeletePush(rid: longint): boolean;

    property      FileName: string
                  read  FFileName
                  write SetFileName;
    property      Flush: FlushType
                  read  FFlush
                  write FFlush;
    end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Len''s', [TVarFile]);
end;

{Methods that set properties}
Procedure  TVarFile.SetFileName(Value: string);

begin
  if FFileName <> Value then
    if Value = '' then Close
    else begin
      AssignFile(F,Value);
      {$I-}  reset(F);  {$I+}
      if IOResult <> 0 then begin
        ShowMessage('The file '+Value+' does not exist.');
        exit;
        end;
      CloseFile(F);
      FFileName := Value;
      FOpen := FALSE;
      end;
   end;



{These routines are useful later }
function TVarFile.FindArea(BlocksWanted: longint): longint;
{This function looks for a group of blocks with BlocksWanted blocks
in the group. It first looks at the free list and if it finds one there
it returns the starting block number. If there is none on the free list
it returns a block at the end of the file.}

var
  i: integer;

begin
  {Assume not on the free list.}
  Result := (FileSize(F)-RecOffset) div P.BlockSize;
  i := 0;
  while i < P.FreeCount do
    if FLP^[i].count = BlocksWanted then begin
      Result := FLP^[i].start;
      {Now remove the entry}
      if i < P.FreeCount-1 then
        move(FLP^[i+1],FLP^[i],8*(P.FreeCount-i-1)); {cover}
      dec(P.FreeCount);
      fillchar(FLP^[P.FreeCount],8,0); {erase last entry}
      exit;
      end
    else if FLP^[i].count > BlocksWanted then begin
           Result := FLP^[i].start;
           FLP^[i].start := FLP^[i].start+BlocksWanted;
           FLP^[i].count := FLP^[i].count-BlocksWanted;
           exit;
           end
         else inc(i);
    end;


procedure TVarFile.AddToFreeList(GivenStart, GivenCount: longint);
{When a group of blocks becomes free it is put on the free list. The
list is kept in order of starting blocks so that when adjacent groups
fall on the list they can be coalesced.}

var
  i,k: integer;

begin
  if P.FreeCount = P.FreeCapacity then exit; {no more room}
  if P.FreeCount = 0 then begin   {Very first entry}
    FLP^[0].start := GivenStart;
    FLP^[0].count := GivenCount;
    inc(P.FreeCount);
    if P.FreeCount > P.MaxFree then P.MaxFree := P.FreeCount;
    exit;
    end;
  i := 0;
  while (i < P.FreeCount) and (FLP^[i].start < GivenStart) do inc(i);
  if i < P.FreeCount then move(FLP^[i],FLP^[i+1],(P.FreeCount-i)*8);
  {Makes a hole in the list for the new entry.}
  FLP^[i].start := GivenStart;
  FLP^[i].count := GivenCount;
  inc(P.FreeCount);
  if P.FreeCount > P.MaxFree then P.MaxFree := P.FreeCount;
  {now coalesce, if possible}
  k := 0;
  if (i>0) and (FLP^[i-1].start+FLP^[i-1].count = FLP^[i].start)
    then k := k + 1;
  if (i<P.FreeCount-1) and (FLP^[i].start+FLP^[i].count = FLP^[i+1].start)
    then k := k + 2;
  case k of
    0: ;
    1: begin {Coalesce with previous entry}
      FLP^[i-1].count := FLP^[i-1].count+FLP^[i].count;
      if i+1 < P.FreeCount then
        move(FLP^[i+1],FLP[i],8*(P.FreeCount-i-1));
      dec(P.FreeCount);
      fillchar(FLP^[P.FreeCount],8,0);
      end;
    2: begin {Coalesce with following entry}
      FLP^[i].count := FLP^[i].count+FLP^[i+1].count;
      if i+2 < P.FreeCount then
        move(FLP^[i+2],FLP^[i+1],8*(P.FreeCount-i-2));
      dec(P.FreeCount);
      fillchar(FLP^[P.FreeCount],8,0);
      end;
    3: begin {Coalesce with both previous and following entries}
      FLP^[i-1].count := FLP^[i-1].count+FLP^[i].count+FLP^[i+1].count;
      if i+2 < P.FreeCount then
        move(FLP^[i+2],FLP^[i],8*(P.FreeCount-i-2));
      dec(P.FreeCount,2);
      fillchar(FLP^[P.FreeCount],16,0);
      end;
    end;
  end;


procedure TVarFile.DoFlush;

begin
  if not FOpen then exit;
  seek(F,0);
  BlockWrite(F,P,32);
  BlockWrite(F,RLP^,8*P.RecCapacity);
  BlockWrite(F,FLP^,8*P.FreeCapacity);
  CloseFile(F);
  reset(F,1);
  end;


{Methods of TVarFile}
constructor TVarFile.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FFileName := '';
  FOpen := FALSE;
  FFlush := Often;
  end;


destructor TVarFile.Destroy;

begin
  Close;
  inherited Destroy;
  end;


procedure TVarFile.Open;

begin
  if FOpen then exit;
  FOpen := FALSE;  {Assume the worst}
  if FFileName = '' then begin
    ShowMessage('No file named yet');
    exit;
    end;
  AssignFile(F,FFileName);
  reset(F,1);
  seek(F,0);
  BlockRead(F,P,32); {Load the prefix into the component.}
  GetMem(RLP,P.RecCapacity*8);
  BlockRead(F,RLP^,P.RecCapacity*8); {Read LocItems into the component.}
  GetMem(FLP,P.FreeCapacity*8);
  BlockRead(F,FLP^,P.FreeCapacity*8); {Same for free list}
  RecOffset := 32 + 8*(P.RecCapacity + P.FreeCapacity);  {Handy}
  FOpen := TRUE;
  end;


procedure TVarFile.Close;

begin
  if not FOpen then exit;
  DoFlush;
  FreeMem(RLP);
  FreeMem(FLP);  {If you GetMem, you FreeMem.}
  CloseFile(F);
  FFileName := '';
  FOpen := FALSE;
  end;


function TVarFile.Exist(rid: longint): boolean;
{Returns TRUE if the record with id rid exists. As an important
side effect, the LI LocItem is filled with the LocItem for the record.}

begin
  if not FOpen then Open;
  Result := TRUE;
  if (rid > P.MaxRec) or (rid <= 0) then Result := FALSE
  else begin
    LI := RLP^[rid-1];
    if LI.count = 0 then Result := FALSE;
    end;
  end;



function TVarFile.GetBufSize(rid: longint): longint;
{Return the number of bytes needed to hold record with id rid.}

begin
  if Exist(rid) then Result := P.BlockSize*LI.count
  else Result := 0;
  end;



function TVarFile.GetARid: longint;
{Returns the rid for the next record to be added to the file.
The rid may be that of a recently deleted record if one has been pushed.
The component maintains a stack in the LocItem array that keeps track
of the rids of records deleted by the DeletePush method.}

begin
  if not FOpen then Open;
  if P.RecCapacity = P.RecCount then begin
    ShowMessage('No room for another record.');
    Result := 0;
    exit;
    end;
  if P.NextRid > 0 then begin  {NextRid > 0 indicates the stack is empty}
    Result := P.NextRid;
    inc(P.NextRid);
    end
  else begin  {NextRid < 0 indicates the stack has entries}
    Result := -P.NextRid;  {The rid of a deleted record.}
    P.NextRid := RLP^[Result-1].start;
    end;
  end;


function TVarFile.Fetch(rid: longint; Buf: pointer): boolean;
{Deliver the record with id rid to the buffer pointed to by Buf.
The buffer will be overwritten if not enough space is allocated.}

begin
  Result := FALSE;
  if not Exist(rid) then exit;
  Result := TRUE;
  seek(F,RecOffset+P.BlockSize*LI.start);
  BlockRead(F,Buf^,LI.count*P.BlockSize);
  end;


function TVarFile.Insert(rid: longint; Buf: pointer;
              RL: longint): boolean;
{This inserts a new record. The parameter rid is presumed to be
the result of a call to GetARid. RL is the number of BYTES in the
record. The routine will compute the number of blocks needed.}

var
  BlocksNeeded: longint;

begin
  if not FOpen then Open;
  if P.RecCapacity = P.RecCount then begin
    ShowMessage('No room for another record.');
    Result := FALSE;
    exit;
    end;
  Result := (rid > 0) and (rid <= P.RecCapacity) and (not Exist(rid));
  if not Result then exit;
  BlocksNeeded := (RL + P.BlockSize - 1) div P.BlockSize;
  LI.start := FindArea(BlocksNeeded);
  LI.count := BlocksNeeded;
  RLP^[rid-1] := LI;
  seek(F,RecOffset + LI.start*P.BlockSize);
  BlockWrite(F,Buf^,LI.count*P.BlockSize);
  inc(P.RecCount);
  if P.MaxRec < P.RecCount then P.MaxRec := P.RecCount;
  if FFlush = Often then DoFlush;
  end;



function TVarFile.Restore(rid: longint; Buf: pointer;
               RL: longint): boolean;
{This restores a record back to the file. rid is the id of the
record and RL is the number of bytes in the record. Buf is a
pointer to the place in memory where the record resides. Do not
use this for inserting NEW records. The record  may not occupy
the same disk space from whence it came.}

var
  BlocksNeeded: longint;

begin
  Result := FALSE;
  if not Exist(rid) then exit;
  Result := TRUE;;
  {free old area}
  AddToFreeList(LI.start,LI.count);
  if RL = 0 then begin
    LI.start := 0; LI.count := 0;
    end
  else begin
    BlocksNeeded := (RL + P.BlockSize -1) div P.BlockSize;
    {get new area}
    LI.start := FindArea(BlocksNeeded);
    LI.count := BlocksNeeded;
    seek(F,RecOffset+P.BlockSize*LI.start);
    BlockWrite(F,Buf^,P.BlockSize*LI.count);
    end;
  RLP^[rid-1] := LI;
  if FFlush = Often then DoFlush;
  end;


function TVarFile.DoDelete(rid: longint; Control: cardinal): boolean;

begin
  Result := FALSE;
  if not Exist(rid) then exit;
  Result := TRUE;
  {free old area}
  AddToFreeList(LI.start,LI.count);
  if Control = 0 then LI.start := 0
  else LI.start := P.NextRid;
  LI.count := 0;
  RLP^[rid-1] := LI;
  if Control = 1 then P.NextRid := -rid;
  dec(P.RecCount);
  if FFlush = Often then DoFlush;
  end;




function TVarFile.Delete(rid: longint): boolean;
{This deletes a record and does not stack the rid for reuse.}

begin
  Result := DoDelete(rid,0);
  end;



function TVarFile.DeletePush(rid: longint): boolean;
{This deletes a record from the disk and pushes its rid onto a stack
for reuse. Do not use rids in other records to reference records
in this file if you use this method.}

begin
  Result := DoDelete(rid,1);
  end;


function TVarFile.PointToPrefix: pointer;
{This returns a pointer to the prefix in memory from which you can
get a lot of information. You also get access to other data such as
LI, RLP, FLP etc. You can also mess up the component real good by
writing the wrong thing in this area.}

begin
  if not FOpen then Open;
  Result := @P;
  end;


end.
